"""Manage binaries requirements (like terraform blob and providers)."""
import os
import stat
import shutil
import re
import subprocess
import tempfile
import zipfile
from pathlib import Path
import requests

from .. import constants as const
from ..logs import logger
from ..utils import error


def search_on_github(repo, minor_version, patch_regex, patch, terraform=True):
    """Search release on github."""
    # Start search from the next incremented minor version
    # Note: the github UI serves the first page if the requested version does not exist
    release = 'v{}{}.0'.format(minor_version[:-1], int(minor_version[-1])+1)
    releases_count = 0
    while True:
        result = requests.get('{}?after={}'.format(const.GITHUB_RELEASES.format(repo), release))
        releases = re.findall(r'<a href=\"/{}/releases/tag/.*\">(.*)</a>'.format(repo), result.text)
        releases_count += len(releases)
        for release in releases:
            if re.match(r'^v{}\.{}$'.format(minor_version, patch if patch else patch_regex),
                        release):
                patch = patch or release.split('.')[-1]
                return patch
            # hard limit or it will takes too long time and terraform versions older
            # than 0.7 do not respect naming convention
            if terraform and any(const.LIMIT_TERRAFORM_VERSION in r for r in releases):
                return None
            if releases_count > const.LIMIT_GITHUB_RELEASES:
                return None
        if len(releases) < 1:
            # no more versions available
            break

        release = releases[-1:][0]
    return None

class Utils:
    """Utility for binary management"""
    def __init__(self, project):
        self.project = project

    @property
    def _tf_binary_cache(self):
        """Where tf binary versions are stored"""
        return Path(self.project.cfg.pyt.get('config.tf_binary_cache'))

    def tf_cached_version(self, version):
        """Cached file binary location"""
        return self._tf_binary_cache / 'versions' / version / 'terraform'

    def tf_download(self, version):
        """Download the wanted version"""
        if not self.tf_cached_version(version).is_file():
            os.makedirs(self.tf_cached_version(version).parent, exist_ok=True)

            # Download and extract in user's home if needed
            logger.warning("Version does not exist locally, downloading it")
            _, tmp_file = tempfile.mkstemp(prefix='terraform-', suffix='.zip')
            get = requests.get(
                f'https://releases.hashicorp.com/terraform/{version}/'
                f'terraform_{version}_{const.PLATFORM_SYSTEM}_{const.ARCH_NAME}.zip',
                stream=True)
            with open(tmp_file, 'wb') as _fd:
                for chunk in get.iter_content(chunk_size=128):
                    _fd.write(chunk)
            with zipfile.ZipFile(tmp_file, 'r') as zip_:
                zip_.extractall(path=self.tf_cached_version(version).parent)
            # Permissions not preserved on extract https://bugs.python.org/issue15795
            os.chmod(self.tf_cached_version(version),
                     stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
            os.remove(tmp_file)

    def update_tf_symlink(self, version):
        """Create local link to cached one"""
        # Doing symlink
        if self.project.path.terraform().is_symlink():
            self.project.path.terraform().unlink()
        if not self.tf_cached_version(version).is_file():
            self.tf_download(version)
        self.project.path.terraform().symlink_to(self.tf_cached_version(version))
        logger.warning("Switch done, current terraform version is %s", version)

    def tf_align_version(self, version):
        """Align the tf binary to the one of the wanted version"""
        regex_version = r'(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<patch>[0-9]+)'
        match = re.match(regex_version, version)
        if not match:
            print(match)
            error('The terraform version seems not correct, '
                  'it should be a version number like "X.Y.Z"')
        # Getting current version
        try:
            pr_ = subprocess.run([self.project.path.terraform(), '-v'],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 check=True)
            current_version = re.match(r'^Terraform v(.+)',
                                       pr_.stdout.decode('ascii')).group(1)
        except FileNotFoundError:
            current_version = 'not installed'

        if current_version == version:
            logger.debug("Terraform is already on version %s", version)
            return

        logger.warning("Current terraform version is %s, switching to version %s",
                       current_version, version)
        self.tf_download(version)
        self.update_tf_symlink(version)


def download_custom_provider(provider_name, provider_version, extension="zip"):
    """Download Terraform custom provider."""
    logger.info("Checking custom Terraform provider '%s' at version '%s'",
                provider_name, provider_version)
    github_base = 'https://github.com/'
    github_endpoint = f"{github_base}{provider_name}"
    supported_extensions = ["zip", "tar.gz", "tar.bz2"]

    if extension not in supported_extensions:
        error(f"Extension {extension} is not supported. "
              f"Only {', '.join(supported_extensions)} are.")

    req = requests.get(github_endpoint)
    if req.status_code != 200:
        error(f"The terraform provider {provider_name} does not exist ({github_endpoint})")

    match = re.match(r'^v?([0-9]+.[0-9]+).?([0-9]+)?(-[a-z]+)?$', provider_version)
    if not match:
        error('The provider version does not seem correct, it should be a version number like '
              '"X.Y", "X.Y.Z" or "X.Y.Z-custom"')
    minor_version, patch, custom = match.groups()

    patch = search_on_github(provider_name, minor_version, r'[0-9]+(-[a-z_-]+)?',
                             patch, terraform=False)

    if not patch:
        error(f"The provider version '{provider_name}-{provider_version}' does not exist")

    full_version = f"{minor_version}.{patch}"
    # Getting current version

    plugins_path = os.path.expanduser(
        f'~/.terraform.d/plugins/{const.PLATFORM_SYSTEM}_{const.ARCH_NAME}')
    os.makedirs(plugins_path, exist_ok=True)
    provider_short_name = provider_name.split('/', 1)[1]
    bin_name = f'{provider_short_name}_{full_version}'
    if provider_version.startswith('v'):
        # Do it now so we keep full_version clean for building url
        bin_name = '{n}_v{v}'.format(n=provider_short_name, v=full_version)
    else:
        bin_name = '{n}_{v}'.format(n=provider_short_name, v=full_version)
    tf_bin_path = os.path.join(plugins_path, f'{provider_short_name}_v{full_version}')

    if not os.path.isfile(tf_bin_path):
        # Download and extract in user's home if needed
        logger.warning("Provider version does not exist locally, downloading it")
        handle, tmp_file = tempfile.mkstemp(prefix='terraform-', suffix="." + extension)
        req = requests.get(
            f'https://github.com/{provider_name}/releases/download/v{full_version}/'
            f'{bin_name}_{const.PLATFORM_SYSTEM}_{const.ARCH_NAME}.{extension}',
            stream=True)
        with open(tmp_file, 'wb') as fd_:
            for chunk in req.iter_content(chunk_size=128):
                fd_.write(chunk)
        shutil.unpack_archive(tmp_file, plugins_path)
        # Permissions not preserved on extract https://bugs.python.org/issue15795
        os.chmod(tf_bin_path,
                 os.stat(tf_bin_path).st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
        os.remove(tmp_file)
        logger.info("Download done, current provider version is %s", full_version)
    else:
        logger.debug("Current provider version is already %s", full_version)



"""All terraform commands, with proper context."""
import os
from copy import deepcopy
import subprocess
import shutil
import json

from ..logs import logger, get_logger
from . import binaries

log = get_logger(__name__, "DEBUG")  # pylint: disable=invalid-name

class Command:
    """Terraform command executor"""

    def __init__(self, project):
        """
        :param project.Project project: the project on which the commands shall be run"""
        self.project = project
        self.utils = binaries.Utils(project)

    @property
    def _tf_bin(self):
        """Terraform binary path"""
        if not self.project.path.terraform().is_file():
            self.utils.tf_align_version(self.project.cfg.pyt.get('config.tf_version'))
        return self.project.path.terraform()
        #if shutil.which("terraform") is not None:
        #    return shutil.which("terraform")

    def update_tf_version(self, version):
        """Update terraform to the wanted version"""
        self.utils.tf_align_version(version)

    def check_tf_version(self):
        """Check tf version and install"""
        if self.project.cfg.stack.get("terraform_version"):
            self.update_tf_version(self.project.cfg.stack["terraform_version"])

    def update_tf_providers(self):
        """Locally install tf providers"""
        # do we need a custom provider ?
        for provider, _config in self.project.cfg.stack.get('custom-providers', {}).items():
            if isinstance(_config, str):
                # This should be the version
                binaries.download_custom_provider(provider, _config)
            else:
                # _config should be a hash of version / extension
                binaries.download_custom_provider(
                    provider, _config['version'], _config['extension'])

    ##  TF WRAPPING       ##

    def list_providers(self):
        """Return list of providers"""
        cmd = f"""find {self.project.path.stack()} -type f -iname "*.tf" -exec awk '
          BEGIN {{print "List of providers and versions:"}}
          / *provider +".*"/ {{
                PROV = $$2;
                while ($$0!~/^[^#]*}}/)
                      {{if ($$0~/^ *version/) match($$0,/version.*= *"(.+)" */,VER);
                       getline;}};
                print "  ", PROV, " is " VER[1]}}' {{}} +
        """
        print(subprocess.Popen(cmd, shell=True).wait())

    def list_modules(self):
        """Return list of modules"""
        cmd = f"""find {self.project.path.stack()} -type f -iname "*.tf" -exec awk '
          BEGIN {{print "List of modules and required sources:"}}
          / *module +".*"/ {{
                USEDBY = USEDBY ", " $$2;
                while ($$0!~/^[^#]*}}/)
                {{if ($$0~/^ *source/) match($$0,/ *source *= *"(.+)" */,SOURCE) ;
                       getline ; }};
                print "  ", SOURCE[1], " used by ", USEDBY }}' {{}} +
        """
        print(subprocess.Popen(cmd, shell=True).wait())

    def local_install(self):
        """Locally install required binaries for stack execution"""
        self.update_tf_providers()

    def version(self):
        """Terraform version wrapper function."""
        return self._run_terraform('version')

    def console(self):
        """Terraform console wrapper function."""
        tf_params, env = self.project.cfg.context_for('console')
        return self._run_terraform('console', tf_params=tf_params, env=env)

    def providers(self):
        """Terraform providers wrapper function."""
        tf_params, env = self.project.cfg.context_for('providers')
        return self._run_terraform('providers', tf_params=tf_params, env=env)

    def plan(self):
        """Terraform plan wrapper function."""
        pipe_plan_command = self.project.cfg.pyt.get('config.pipe_plan_command')
        tf_params, env = self.project.cfg.context_for('plan')
        return self._run_terraform('plan', tf_params=tf_params, env=env, pipe=pipe_plan_command)

    def init(self):
        """Terraform init wrapper function."""
        tf_params, env = self.project.cfg.context_for('init')
        tf_params.extend(['-input=false',
                          '-force-copy',
                          '-lock=true',
                          #'-upgrade=true',
                          '-get-plugins=true',
                          '-verify-plugins=true'])
        if self.project.cfg.pyt.get('state.backend', {}).get('s3', {}):
            tf_params.append('-backend=true')
            tf_params.extend(self.project.cfg.stack.backend_setup)
        return self._run_terraform('init', tf_params=tf_params, env=env)

    def _run_terraform(self, action, tf_params=None, env=None, pipe=None):
        """Run Terraform command."""
        # support for custom parameters
        command = [self._tf_bin, action]
        if tf_params is not None:
            if tf_params and tf_params[0] == '--':
                tf_params = tf_params[1:]
            command += tf_params

        cmd_env = env if env else deepcopy(os.environ)

        with subprocess.Popen(command, cwd=self.project.path.stack(),
                              env=cmd_env, shell=False,
                              stdout=subprocess.PIPE if pipe else None) as process:
            logger.debug('Execute command "%s"', command)
            log.info("Running command: '%s'", ' '.join([str(x) for x in command]))
            log.info("On path '%s'", self.project.path.stack())
            #log.info('And with env: %s', {x: cmd_env[x] for x in sorted(dict(cmd_env))})
            log.info('And with env: %s', json.dumps(dict(cmd_env), indent=2, sort_keys=True))
            if pipe:
                logger.debug('Piping command "%s"', pipe)
                with subprocess.Popen(pipe, cwd=self.project.path.stack(),
                                      env=cmd_env, shell=False,
                                      stdin=process.stdout) as pipe_process:
                    try:
                        pipe_process.communicate()
                    except KeyboardInterrupt:
                        logger.warning('Received Ctrl+C')
                    except:  # noqa
                        pipe_process.kill()
                        pipe_process.wait()
                        raise
                    pipe_process.poll()
            try:
                process.communicate()
            except KeyboardInterrupt:
                logger.warning('Received Ctrl+C')
            except:  # noqa
                process.kill()
                process.wait()
                raise
            return process.poll()

    def run(self):
        """Execute the command asked for by the cli input"""
        action = getattr(self, self.project.input.args.get('subcommand'), None)
        if action is None:
            return self._run_terraform(self.project.input.args.get('subcommand'))
        logger.info("Going to run command '%s'", self.project.input.args.get('subcommand'))
        return action()


#def terraform_apply(wrapper_config):
#    """Terraform apply wrapper function."""
#    always_trigger_init = wrapper_config['config'].get('always_trigger_init', False)
#    logger.debug("Checking 'always_trigger_init' option: {}".format(always_trigger_init))
#    if always_trigger_init:
#        logger.info('Init has been activated in config')
#        terraform_init(wrapper_config)
#
#    # do not force plan if unsafe
#    if wrapper_config['unsafe']:
#        return run_terraform('apply', wrapper_config)
#    else:
#        # plan config
#        plan_path = '{}/.run/plan_{}'.format(wrapper_config['rootdir'],
#            ''.join(random.choice(string.ascii_letters) for x in range(10)))
#        plan_wrapper_config = deepcopy(wrapper_config)
#        plan_wrapper_config['tf_params'][1:1] = ['-out', plan_path]
#        plan_return_code = run_terraform('plan', plan_wrapper_config)
#
#        # return Terraform return code if plan fails
#        if plan_return_code > 0:
#            return plan_return_code
#
#        # ask for confirmation
#        colored_account = colored(plan_wrapper_config['account'], 'yellow')
#        colored_environment = colored(plan_wrapper_config['environment'], 'red')
#        colored_region = colored(plan_wrapper_config['region'], 'blue')
#        colored_stack = colored(plan_wrapper_config['stack'], 'green')
#
#        if plan_wrapper_config['environment'] == 'global':
#            env_msg = '''
#    Account : {}
#Environment : {}
#      Stack : {}
#'''.format(colored_account, colored_environment, colored_stack)
#        else:
#            env_msg = '''
#    Account : {}
#Environment : {}
#     Region : {}
#      Stack : {}
#'''.format(colored_account, colored_environment, colored_region, colored_stack)
#
#        print('\nDo you really want to apply this plan on the following stack ?\n',
#              env_msg)
#        apply_input = input("'yes' to confirm: ")
#
#        try:
#            if apply_input == 'yes':
#                # apply config
#                apply_wrapper_config = deepcopy(wrapper_config)
#                apply_wrapper_config['tf_params'].append(plan_path)
#                apply_return_code = run_terraform('apply', apply_wrapper_config)
#
#                return apply_return_code
#            else:
#                logger.warning('Aborting apply.')
#        finally:
#            # delete plan
#            os.remove(plan_path)
#
#
#def terraform_console(wrapper_config):
#    """Terraform console wrapper function."""
#    return run_terraform('console', wrapper_config)
#
#
#def terraform_destroy(wrapper_config):
#    """Terraform destroy wrapper function."""
#    return run_terraform('destroy', wrapper_config)
#
#
#def terraform_fmt(wrapper_config):
#    """Terraform fmt wrapper function."""
#    return run_terraform('fmt', wrapper_config)
#
#
#def terraform_force_unlock(wrapper_config):
#    """Terraform force-unlock wrapper function."""
#    return run_terraform('force-unlock', wrapper_config)
#
#
#def terraform_get(wrapper_config):
#    """Terraform get wrapper function."""
#    # force update
#    if not any('-update' in x for x in wrapper_config['tf_params']):
#        wrapper_config['tf_params'][1:1] = ['-update']
#
#    # call subcommand
#    return run_terraform('get', wrapper_config)
#
#
#def terraform_graph(wrapper_config):
#    """Terraform graph wrapper function."""
#    return run_terraform('graph', wrapper_config)
#
#
#def terraform_import(wrapper_config):
#    """Terraform import wrapper function."""
#    return run_terraform('import', wrapper_config)
#
#
#def terraform_output(wrapper_config):
#    """Terraform output wrapper function."""
#    return run_terraform('output', wrapper_config)
#
#
#def terraform_plan(wrapper_config):
#    """Terraform plan wrapper function."""
#    always_trigger_init = wrapper_config['config'].get('always_trigger_init', False)
#    logger.debug("Checking 'always_trigger_init' option: {}".format(always_trigger_init))
#    if always_trigger_init:
#        logger.info('Init has been activated in config')
#        terraform_init(wrapper_config)
#    return run_terraform('plan', wrapper_config)
#
#
#def terraform_providers(wrapper_config):
#    """Terraform providers wrapper function."""
#    return run_terraform('providers', wrapper_config)
#
#
#def terraform_refresh(wrapper_config):
#    """Terraform refresh wrapper function."""
#    return run_terraform('refresh', wrapper_config)
#
#
#def terraform_show(wrapper_config):
#    """Terraform show wrapper function."""
#    return run_terraform('show', wrapper_config)
#
#
#def terraform_state(wrapper_config):
#    """Terraform state wrapper function."""
#    return run_terraform('state', wrapper_config)
#
#
#def terraform_taint(wrapper_config):
#    """Terraform taint wrapper function."""
#    return run_terraform('taint', wrapper_config)
#
#
#def terraform_untaint(wrapper_config):
#    """Terraform untaint wrapper function."""
#    return run_terraform('untaint', wrapper_config)
#
#
#def terraform_validate(wrapper_config):
#    """Terraform validate wrapper function."""
#    return run_terraform('validate', wrapper_config)



"""A terraform smart wrapper.

This script should let run terraform everywhere in a consistent way.
"""
import os
import sys
import logging
from copy import deepcopy

from . import constants as const
from .logs import logger
from . import config
from . import session
from . import project


def main():
    """Execute pyterraform wrapper."""
    stack = project.Project()
    return stack.run()


if __name__ == "__main__":
    main()



"""Common project's constants"""
from pathlib import Path
import platform


def get_architecture():
    """Get system architecture name normalized for terraform."""
    platform_system = platform.machine()
    if 'arm' in platform_system:
        return 'arm'
    if platform_system in ('x86_64', 'darwin'):
        return 'amd64'
    return '386'

RC_OK = 0
RC_KO = 1
RC_UNK = 2

LIMIT_TERRAFORM_VERSION = 'v0.11.1'
LIMIT_GITHUB_RELEASES = 42
GITHUB_RELEASES = 'https://github.com/{}/releases'
ARCH_NAME = get_architecture()
PLATFORM_SYSTEM = platform.system().lower()

HOME_DIR = Path.home()
CONF_DIR = Path('pyterraform')

CWD = Path.cwd()



"""Common path position.
"""
import sys
from pathlib import Path
import itertools

from .logs import logger
from .utils import error
from . import constants as const


class Conf:
    """Pyterraform configuration directory"""
    def __init__(self, paths):
        self.paths = paths

    def __call__(self):
        return self.paths.root() / const.CONF_DIR

    def pyterraform(self):
        """Configuration file for tfwrapper (YAML)"""
        return self() / 'pyterraform.yml'

    def state(self):
        """Configuration file for state backends (YAML)"""
        return self() / 'state.yml'

class Stack:
    """Stack directory"""
    def __init__(self, paths):
        self.paths = paths

    def __call__(self):
        if not self.paths._cache.get('stack'):
            self.paths._cache['stack'] = self.paths.get_stack_path(const.CWD)
        return self.paths._cache['stack']

    def config(self):
        """Stack config file as:
        - <root_dir>/<stack/path/definition>/stack.yml"""
        return self() / 'stack.yml'


class Paths:
    """Project relevant paths.
    The following structure is expected:
    /pyterraform:    global options of the project and the wrapper
    /<stack>/<env>/: proper TF stack (or what specified into pyterraform conf)
    /modules/:       shared modules cacheing
    /terraform:      binary of terraform
    /.run:           runtime files"""

    def __init__(self, project):
        self.project = project
        self._cache = dict()

    def root(self):
        """Wrapper root folder"""
        if not self._cache.get("root"):
            for i in range(0, 5):
                if (Path("../" * i) / const.CONF_DIR).is_dir():
                    self._cache["root"] = Path("../" * i).absolute().resolve()
                    logger.debug("Detected root folder at '%s'",
                                 self._cache["root"])
                    break
            else:
                error("Cannot locate project root folder, are you inside it?")
                sys.exit(const.RC_KO)
        return self._cache["root"]

    @property
    def conf(self):
        """Pyterraform configuration directory (under root path)"""
        return Conf(self)

    def terraform(self):
        """Terraform binary"""
        return self.root() / "terraform"

    def _get_stackinfo_from_cwd(self, path=None):
        """Find, if any, the stack name from the working directory"""
        path = Path(path) if path else Path.cwd()
        try:
            children = path.relative_to(self.root()).parts
            meta = dict(itertools.zip_longest(self.project.cfg.pyt.stack_folder_structure,
                                              children))
        except ValueError:
            logger.info("Terraform is not running from root project folder"
                        "Working dir is: %s - root dir is %s",
                        path, self.root())
        return meta

    def get_stack_path(self, path=None):
        """Compute from cwd the stack path"""
        meta = self._get_stackinfo_from_cwd(path)
        path = Path()
        for dir_ in self.project.cfg.pyt.stack_folder_structure:
            path = path / meta[dir_]
        return self.root() / path

    @property
    def stack(self):
        """Stack path as, usually, <root>/<stack>/<env>.
        Such path could be defined from cli args or from cwd"""
        return Stack(self)

    def modules(self):
        """Terraform modules folder"""
        return self.root() / "modules"

    def run(self):
        """Execution data"""
        dir_ = self.root() / '.run'
        if not dir_.is_dir():
            dir_.mkdir(exist_ok=True)
        return self.root() / '.run'



"""Configuration object storage for pyterraform and stacks"""
from abc import ABC, abstractmethod
import sys
from pathlib import Path
import yaml
from schema import Schema, Optional, Or, SchemaError
from .. import constants as const
from ..logs import logger


# pylint: disable=too-few-public-methods,missing-function-docstring
class Setups(ABC):
    """Generic setup placeholder"""

    def __init__(self, project):
        self._data = None
        self.project = project

    @property
    def data(self):
        """Raw data (please don't change)"""
        if self._data is None:
            self.load_data()
        return self._data
    def _get(self, path, default=None):
        data = self.data
        try:
            for item in path.split('.'):
                data = data[item]
            return data
        except KeyError:
            return default
    def _set(self, path, value):
        data = self.data
        for key in path.split('.')[:-1]:
            data = data.setdefault(key, dict())
        data[path.split('.')[-1]] = value
    def get(self, path, default=None):
        """Extract wanted value"""
        return self._get(path, default)
    @abstractmethod
    def load_data(self):
        """Load data from proper sources"""
        return


class Pyterraform(Setups):
    """Set up of pyterraform wrapper"""
    def load_data(self):
        """Load data from state and conf, interpolate with args"""
        state = self._load_state()
        config = self._load_config()
        self._data = {'state': state,
                      'config': config}

    @property
    def stack_folder_structure(self):
        """Structure of folder"""
        return self._get('config.folder_structure').split('.')
    @property
    def _state_schema(self):
        """Schema validation and defaults"""
        return Schema( \
    {Optional('profile'): str,
     Optional('region'): str,
     Optional('assume_role'): str,
     Optional('backend'): {Optional('s3'): {
         'bucket': str,
         'key': str,
         'dynamodb_table': str,
         Optional('acl', default='private'): str}}})

    def _load_state(self):
        """Load state example"""
        if not self.project.path.conf.state().is_file():
            logger.warning("No state configuration file found!")
            return dict()
        with self.project.path.conf.state().open() as _f:
            logger.debug("Loading state config from '%s'", self.project.path.conf.state())
            config = yaml.safe_load(_f)
        try:
            state = self._state_schema.validate(config)
            return state
        except SchemaError as ex:
            logger.error('Configuration error in %s : %s',
                         self.project.path.conf.state(), ex)
            sys.exit(const.RC_KO)

    @property
    def _config_schema(self):
        """Schema validation and defaults"""
        return Schema({ \
    Optional('always_trigger_init', default=False): bool,
    Optional('pipe_plan_command', default='cat'): str,
    Optional('folder_structure', default='stack.environment'): str,
    Optional('tf_version', default='0.12.21'): str,
    Optional('tf_binary_cache', default=Path.home() / '.terraform' / 'binaries'): str,})
#    Optional('tf_plugin_dir', default='/tmp/terraform.d/plugin'): str,
#    Optional('tf_data_dir', default='/tmp/terraform.d/data/{stack}/{environment}'): str})

    def _load_config(self):
        """Load config example"""
        try:
            with self.project.path.conf.pyterraform().open() as _f:
                logger.debug("Loading pyterraform wrapper config from '%s'",
                             self.project.path.conf.pyterraform())
                config = yaml.safe_load(_f)
        except FileNotFoundError:
            logger.warning("No pyterraform configuration file found!")
            config = dict()
        try:
            config = self._config_schema.validate(config)
            return config
        except SchemaError as ex:
            logger.error('Configuration error in %s : %s',
                         self.project.path.conf.pyterraform(), ex)
            sys.exit(const.RC_KO)

    #@property
    #def plugin_cache_dir(self):
    #    """Where to cache tf plugins"""
    #    if not self._get('plugin_cache_dir'):
    #        self._set('plugin_cache_dir',
    #                  self.src.args.get('plugin_cache_dir') or \
    #                  self.src.environment.get('TF_plugin_cache_dir') or \
    #                  self.src.ftf.get('plugin_cache_dir'))
    #    return self._get('plugin_cache_dir')


class Stack(Setups):
    """Setting of a single stack"""
    def load_data(self):
        """Load data from state and conf, interpolate with args"""
        config = self._load_config()
        self._data = config

    @property
    def validation_schema(self):
        """Validate schema structure"""
        return Schema( \
        {
            #Optional('state_configuration_name'): str,
            #Optional('aws'): {
            #    'general': {
            #        'account': str,
            #        'region': str
            #    },
            #    'credentials': {
            #        'profile': str,
            #    }},
            Optional('vars'): {str: str},
            Optional('var-file'): str,
            Optional('terraform'): {
                Optional('custom-providers'): {str: Or(str, {'version': str, 'extension': str})}}
        })


    def _load_config(self):
        """Read stack configuration file, merged with element implicit into the cwd"""
        try:
            with self.project.path.stack.config().open() as _f:
                stack_config = yaml.safe_load(_f)
        except FileNotFoundError:
            logger.warning("No stack configuration found!")
            stack_config = dict()
        try:
            stack_config = self.validation_schema.validate(stack_config)
            stack_config.update(self.project.input.path)
            return stack_config
        except SchemaError as ex:
            logger.error('Configuration error in %s : %s',
                         self.project.path.stack.config(), ex)
            sys.exit(const.RC_KO)

    @property
    def backend_setup(self):
        """Option for backend setup"""
        params = list()
        for key, value in \
                self.project.cfg.pyt.get('state.backend', {}).get('s3', {}).items():
            value = value.format(**self.data)
            params.extend(['-backend-config', f"{key}={value}"])
            #params.append(f'-backend-config="{key}={value}"')
        return params



"""A terraform project.
"""
import sys

from . import paths
from .config import Configuration
from . import inputs
from . import session
from . import constants as const
from .terraform import Command
from .logs import set_root_logger, get_logger, logger

log = get_logger(__name__, 'DEBUG')  # pylint: disable=invalid-name


class Project:  # pylint: disable=too-few-public-methods
    """A terraform project"""

    def __init__(self):
        self.path = paths.Paths(self)
        self.input = inputs.Data(self)
        self.cfg = Configuration(self)
        self.tf = Command(self)  # pylint: disable=invalid-name
        self.session = session.Session(self)
        self.enrich_logging()

    def enrich_logging(self):
        """Based on cli inputs, enrich logging"""
        if self.input.args.get('log_to_file') or self.input.environment.get('log_to_file') \
                or self.cfg.pyt.get('config.log_to_file'):
            set_root_logger(log_to_file=self.path.run() / "pyterraform.logs")
            logger.info("Enabled logging to file \"%s\"", self.path.run() / "pyterraform.logs")
            log.info("Enabled log to file for verbose analysis")

    def run(self):
        """Execute the command as request by cli input"""
        returncode = None
        # check terraform version
        #if args.subcommand not in ('foreach', 'providers', 'switchver', 'version'):
        #    self.tf.check_tf_version()
        #if args.subcommand in ["init", "bootstrap"]:
        #    self.tf.update_tf_providers()

        # run terraform finally!
        if self.cfg.pyt.get('config.tf_data_dir'):
            logger.info("Plan data will be cached on %s",
                        self.cfg.pyt.get('config.tf_data_dir').format(**self.cfg.stack.data))
        returncode = self.tf.run()
        log.info("The exit status is '%s'", returncode)
        if returncode is not None:
            sys.exit(returncode)
        else:
            sys.exit(const.RC_OK)



"""Logging setup"""
import os
import logging
import colorlog


# pylint: disable=invalid-name
LOG_FORMAT = "{log_color}{levelname: <7}{reset} {purple}{name}{reset}: {bold}{message}{reset}"
handler = colorlog.StreamHandler()
handler.setFormatter(colorlog.ColoredFormatter(LOG_FORMAT, style='{'))
logger = colorlog.getLogger('tfwrapper')
logger.addHandler(handler)
logger.setLevel(logging.INFO)



def set_root_logger(log_to_file=None, log_to_stream=None):
    """Set root logger for more verbose analisys"""
    rootlog = logging.getLogger()
    handlers = getattr(rootlog, 'handlers', list())
    formatter = logging.Formatter(
        fmt=set_root_logger.message_format,
        datefmt=set_root_logger.date_format)
    if log_to_file:
        import logging.handlers as handlers_  # pylint: disable=import-outside-toplevel
        file_ = handlers_.RotatingFileHandler(log_to_file, backupCount=10, maxBytes=20*2**100)
        file_.doRollover()
        handlers.append(file_)
    if log_to_stream:  # like sys.stdout
        handlers.append(logging.StreamHandler(log_to_stream))
    for _handler in handlers:
        _handler.setFormatter(formatter)

set_root_logger.message_format = (
    '%(asctime)s.%(msecs)03d %(process)d-%(thread)d %(name)-8s:'
    '[%(levelname)6s]:[%(filename)s:%(lineno)d]: %(message)s')
set_root_logger.date_format = '%Y-%m-%d %H:%M:%S'

# Logger Factory
def get_logger(logger_name=None, level=None):
    """Generate a logger.

    Usage:
    >>> logger = get_logger()
    >>> logger = get_logger(__name__, "DEBUG")

    parameters
    ----------
    :param str logger_name: the name of the logger, if none the root logger will be used
    :param str level: the logger level, if not set it will be user DEBUG or INFO according to the
        DEBUG_MODE environment variable
    :return logging:
    """
    # Get the requested logger
    logger_ = logging.getLogger(logger_name)

    # Set the logger level
    level = getattr(logging, os.environ.get('LOG_LEVEL', 'NOTDEF'),
                    logging.INFO if not level else getattr(logging, level, logging.INFO))
    logger_.setLevel(level)

    return logger_

if os.environ.get("VERBOSE_AWS"):
    # pylint: disable=invalid-name,missing-docstring,too-few-public-methods
    class OnlySessionInfos(logging.Filter):
        def filter(self, record):
            msg = record.getMessage()
            return msg.startswith("Response body") or msg.startswith("Making request for")

    boto_parser = logging.getLogger("botocore.parsers")
    boto_endpoint = logging.getLogger("botocore.endpoint")
    boto_parser.setLevel(logging.DEBUG)
    boto_endpoint.setLevel(logging.DEBUG)
    boto_parser.addFilter(OnlySessionInfos())
    boto_endpoint.addFilter(OnlySessionInfos())



"""Parser for cli inputs"""
import sys
import argparse
from . import constants as const


def parse_args(project, args):  # pylint: disable=too-many-locals,too-many-statements
    """Parse command line arguments."""
    # pylint: disable=unused-variable
    # terraform params doc
    tf_params_help = 'Any Terraform parameters after a "--" delimiter'

    # argparse
    parser = argparse.ArgumentParser(prog="pyterraform",
                                     description='Terraform wrapper.')
    parser.add_argument("-d", "--debug",
                        action='store_true', default=False,
                        help="Enable debug output.")
    parser.add_argument(
        '-c', '--confdir',
        help='Configuration directory. Used to detect the project root. Defaults to conf.',
        default='conf')
    for stack_element in project.cfg.pyt.stack_folder_structure:  # at least stack and environment
        parser.add_argument(f'--{stack_element}',
                            help='Target stack definition. Autodetected if none is provided.',
                            nargs='?')
    #parser.add_argument('-a', '--account',
    #                    help='Target account. Autodetected if none is provided.',
    #                    nargs='?')
    #parser.add_argument('-r', '--region',
    #                    help='Target region. Autodetected if none is provided.',
    #                    nargs='?')
    #parser.add_argument('--profile',
    #                    help='Target profile. Autodetected if none is provided.',
    #                    nargs='?')
    parser.add_argument('--log-to-file', help='Enable logging to local file.',
                        action='store_true', default=False)
    parser.add_argument('--unattended', help='For automatic run (like CI).',
                        action='store_true', default=False)
    parser.add_argument('-p', '--plugin-cache-dir', help='Plugins cache directory.',
                        default=f'{const.HOME_DIR}/.terraform.d/plugin-cache')

    subparsers = parser.add_subparsers(dest='subcommand',
                                       help='terraform subcommands plus some pyterraform gotchas')

    subparsers.add_parser('list_modules', help='List modules used by terraform')
    subparsers.add_parser('list_providers',
                          help='List providers declared into terraform stack')
    subparsers.add_parser('local_install', help='Install dependencies required by terraform')
    parser_apply = subparsers.add_parser('apply', help='terraform apply')
    #parser_apply.add_argument('-u', '--unsafe',
    #                          help='Do not force plan and human interaction before apply.',
    #                          action='store_true', default=False)
    #parser_apply.add_argument("-l", "--pipe-plan",
    #                          action='store_true', default=False,
    #                          help=("Pipe plan output to the command set in config"
    #                          " or passed in --pipe-plan-command argument (cat by default)."))
    #parser_apply.add_argument("--pipe-plan-command",
    #                          action='store', nargs='?',
    #          help="Pipe plan output to the command of your choice set as argument inline value.")
    parser_apply.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_console = subparsers.add_parser('console', help='terraform console')
    parser_console.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    #parser_destroy = subparsers.add_parser('destroy', help='terraform destroy')
    #parser_destroy.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_fmt = subparsers.add_parser('fmt', help='terraform fmt')
    parser_fmt.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_force_unlock = subparsers.add_parser('force-unlock', help='terraform force-unlock')
    parser_force_unlock.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_get = subparsers.add_parser('get', help='terraform get')
    parser_get.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    #parser_graph = subparsers.add_parser('graph', help='terraform graph')
    #parser_graph.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_import = subparsers.add_parser('import', help='terraform import')
    parser_import.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_init = subparsers.add_parser('init', help='terraform init')
    parser_init.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_output = subparsers.add_parser('output', help='terraform output')
    parser_output.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_plan = subparsers.add_parser('plan', help='terraform plan')
    parser_plan.add_argument("-l", "--pipe-plan",
                             action='store_true', default=False,
                             help=("Pipe plan output to the command set in config"
                                   " or passed in --pipe-plan-command argument (cat by default)."))
    #parser_plan.add_argument("--pipe-plan-command",
    #                         action='store', nargs='?',
    #          help="Pipe plan output to the command of your choice set as argument inline value.")
    parser_plan.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_providers = subparsers.add_parser('providers', help='terraform providers')
    parser_providers.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_refresh = subparsers.add_parser('refresh', help='terraform refresh')
    parser_refresh.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_show = subparsers.add_parser('show', help='terraform show')
    parser_show.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_state = subparsers.add_parser('state', help='terraform state')
    parser_state.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_taint = subparsers.add_parser('taint', help='terraform taint')
    parser_taint.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_untaint = subparsers.add_parser('untaint', help='terraform untaint')
    parser_untaint.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_validate = subparsers.add_parser('validate', help='terraform validate')
    parser_validate.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    parser_version = subparsers.add_parser('version', help='terraform version')
    parser_version.add_argument('tf_params', nargs=argparse.REMAINDER, help=tf_params_help)

    #parser_bootstrap = subparsers.add_parser('bootstrap', help='bootstrap configuration')
    #parser_bootstrap.add_argument('template', nargs='?',
    #        help='template to use during bootstrap', default=None)

    #parser_foreach = subparsers.add_parser('foreach', help='execute command for each stack')
    #parser_foreach.add_argument('-c', dest='shell', action="store_true",
    #          help='execute command in a shell')
    #parser_foreach.add_argument('command', nargs=argparse.REMAINDER,
    #          help='command to execute after a "--" delimiter')

    #parser_switchver = subparsers.add_parser('switchver', help='switch terraform version')
    #parser_switchver.add_argument('version', nargs=1, help='terraform version to use')

    parsed_args = parser.parse_args(args)

    if getattr(parsed_args, 'subcommand') is None:
        parser.print_help(file=sys.stderr)
        raise SystemExit(0)

    #if parsed_args.func == foreach:
    #    if len(parsed_args.command) > 0 and parsed_args.command[0] == '--':
    #        parsed_args.command = parsed_args.command[1:]
    #    if len(parsed_args.command) < 1:
    #        raise ValueError("foreach: error: a command is required")
    #    if parsed_args.shell and len(parsed_args.command) > 1:
    #raise ValueError("foreach: error: -c must be followed by a single argument (hint: use quotes)")
    #    parsed_args.executable = os.environ.get("SHELL", None) if parsed_args.shell else None

    return parsed_args



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyterraform",
    version="0.0.1",
    author="Riccardo Scartozzi <risca>",
    author_email="info@risca.eu",
    description="A wrapper to terraform tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="terraform",
    url="https://github.com/terraform-great-modules/pyterraform",
    packages=setuptools.find_packages(),
    #entry_points={
    #    'console_scripts': [
    #        'my_project = my_project.__main__:main'
    #    ]
    #}
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "Operating System :: POSIX :: Linux"
    ],
    python_requires='>=3.6',
    install_requires=['boto3',
                      'colorlog',
                      'schema',
                      'pyyaml',
                      'requests',
                      'deep_merge']
)



"""Inputs placeholder."""
import os
import sys
import logging
from . import logger
from . import constants as const
from .cli_tools import parse_args

class Data:
    """Data configuration placeholder for:
    - cli arguments;
    - os environment
    - default paths
    """
    def __init__(self, project):
        self.project = project
        self._data = {'args': None,
                      'path': None,
                      'env': dict(os.environ)}

    @property
    def args(self):
        """Store cli arguments"""
        if self._data['args'] is None:
            args = sys.argv[1:]
            logger.debug("Command inputs: %r", args)
            self._data['args'] = vars(parse_args(self.project, args))
            if self._data['args'].get("debug"):
                logger.setLevel(logging.DEBUG)

        return self._data['args']
    @property
    def path(self):
        """Infos from cwd"""
        if self._data['path'] is None:
            self._data['path'] = self.project.path._get_stackinfo_from_cwd(const.CWD)  # pylint: disable=protected-access
        return self._data['path']
    @property
    def environment(self):
        """Environment configuration"""
        return self._data['env']



"""Manage aws sessions"""
import os
import time
import pickle
import sys

import boto3
import botocore

from . import constants as const
from .logs import logger

# pylint: disable=fixme

class Session:
    """AWS session caching and setting"""

    def __init__(self, project):
        self._credentials = None
        self.project = project

    @property
    def profile(self):
        """Return the profile to be used, looking up to:
        - cli
        - state.yml
        - environment"""
        if self.project.cfg.pyt.get("config.profile"):
            return self.project.cfg.pyt.get('config.profile')
        if self.project.cfg.pyt.get('state.profile'):
            return self.project.cfg.pyt.get('state.profile')
        return os.environ['AWS_PROFILE']

    @property
    def session_cache_file(self):
        """Temporary store credential"""
        return self.project.path.run() / f'session_cache_{self.profile}.pickle'
        #return self.project.path.run / f'session_cache_{self.account}_{self.profile}.pickle'

    def _get_session(self):
        """Get or create boto cached session."""
        if self.session_cache_file.is_file() and \
                time.time() - os.stat(self.session_cache_file).st_mtime < 2700:
            with open(self.session_cache_file, 'rb') as _f:
                session_cache = pickle.load(_f)
            session = boto3.Session(aws_access_key_id=session_cache['credentials'].access_key,
                                    aws_secret_access_key=session_cache['credentials'].secret_key,
                                    aws_session_token=session_cache['credentials'].token,
                                    region_name=session_cache['region'])
        else:
            session_args = {"profile_name": self.profile}
            if 'region' in self.project.cfg.pyt.get('state'):
                session_args['region_name'] = self.project.cfg.pyt.get('state.region')
            # TODO: add assume role
            try:
                session = boto3.Session(**session_args)
            except botocore.exceptions.ProfileNotFound:
                logger.error("Profile not found.")
                logger.error("No valid AWS session found. Exiting...")
                sys.exit(const.RC_KO)
            try:
                session_cache = {'credentials': session.get_credentials().get_frozen_credentials(),
                                 'region': session.region_name}
            except botocore.exceptions.ParamValidationError:
                logger.error('Error validating authentication. Maybe the wrong MFA code ?')
                sys.exit(const.RC_KO)
            except Exception:  # pylint: disable=broad-except
                logger.exception('Unknown error')
                sys.exit(const.RC_UNK)
            #with os.fdopen(os.open(self.session_cache_file, os.O_WRONLY | os.O_CREAT,
            #                       mode=0o600), 'wb') as _f:
            os.makedirs(self.project.path.run(), exist_ok=True)
            with open(self.session_cache_file, 'wb') as _f:
                pickle.dump(session_cache, _f, pickle.HIGHEST_PROTOCOL)
        return session

    @property
    def credentials(self):
        """Retun AWS credentials"""
        if not self._credentials:
            self._credentials = self._get_session().get_credentials().get_frozen_credentials()
        return self._credentials

    @property
    def access_key(self):
        """AWS_ACCESS_KEY_ID"""
        return self.credentials.access_key
    @property
    def secret_key(self):
        """AWS_SECRET_ACCESS_KEY"""
        return self.credentials.secret_key
    @property
    def token(self):
        """AWS_SESSION_TOKEN"""
        return self.credentials.token

    def infect_environment(self):
        """Infect current environment with access attributes"""
        os.environ['AWS_ACCESS_KEY_ID'] = self.access_key
        os.environ['AWS_SECRET_ACCESS_KEY'] = self.secret_key
        if self.token:
            os.environ['AWS_SESSION_TOKEN'] = self.token



"""Manage configuration options, getting metadata from filesystem,
current working directory and cli options"""
import os
import sys
import copy

from . import setup


def overwrite_notnull(v1, v2, **kwargs):  # pylint: disable=invalid-name,unused-argument
    """
    Completely overwrites one value with another, if not None.
    """
    if v2 is None:
        return copy.deepcopy(v1)
    return copy.deepcopy(v2)


class Configuration:
    """Configuration metadata holder and parameters interpolation.
    It will manage stack configuration, pyterraform wrapper options
    and environment variables, more..."""

    def __init__(self, project):
        self.project = project
        self.stack = setup.Stack(project)
        self.pyterraform = setup.Pyterraform(project)

    @property
    def pyt(self):
        """Alias to pyterraform"""
        return self.pyterraform

    @property
    def _data(self):
        """Just for convenience"""
        return self.project.data

    #@property
    #def merged_variables(self):
    #    """Return all variables merged together.
    #    Order (by priority):
    #    - cli_args
    #    - path_info
    #    - os.environ
    #    - stack
    #    - state
    #    - tf_wrapper
    #    """
    #    mixin = dict()
    #    priority = [self.args, self.path, dict(os.environ), self.stack,
    #                self.state, self.tf]
    #    kpolicy = dict(merge_lists=overwrite_notnull, merge_ints=overwrite_notnull,
    #                   merge_floats=overwrite_notnull, merge_strings=overwrite_notnull,
    #                   merge_other=overwrite_notnull)
    #    deep_merge.merge(mixin, *reversed(priority), **kpolicy)
    #    return mixin

    #@staticmethod
    #def infec
    def get_tf_env(self):
        """Environment option for terraform"""
        envs = dict()
        cli_args = list()
        if self.pyt.get('config.tf_data_dir'):
            envs['TF_DATA_DIR'] = self.pyt.get('config.tf_data_dir').format(**self.stack.data)
        if self.pyt.get('config.tf_plugin_dir'):
            cli_args.append(f"-plugin-dir={self.pyt.get('config.tf_plugin_dir')}")
        if cli_args:
            envs['TF_CLI_ARGS'] = ' '.join(cli_args)
        for var, value in envs.items():
            if value is not None:
                os.environ[var] = str(value)

    def get_stack_custom_env(self):
        """Custom runtime env. Is it needed?"""
        terraform_vars = dict()
        terraform_vars['environment'] = self.stack.get("environment")
        terraform_vars['stack'] = self.stack.get("stack")
        # AWS variables
        terraform_vars['aws_access_key'] = self.project.session.access_key
        terraform_vars['aws_secret_key'] = self.project.session.secret_key
        terraform_vars['aws_token'] = self.project.session.token
        return terraform_vars

    def get_stack_tfvariables(self):
        """Return dict of variable to be passed to TF as environment."""
        tf_vars = dict()
        for key, value in self.stack.get("vars", dict()).items():
            if value in [None, ""]:
                continue
            tf_vars[f'TF_VAR_{key}'] = value
        return tf_vars

    def get_stack_varfile(self):
        """Return the variable file to be passed to TF"""
        vfile = self.stack.get('var-file')
        if vfile:
            return f'-var-file={vfile}'
        return ''

    def context_for(self, command):
        """Extract the proper execution contest for the given command.
        :return: cli arguments definition and environment
        :rtype: list(str), dict"""
        #TF_IN_AUTOMATION input=False
        cli_args = list()
        envs = copy.deepcopy(self.project.input.environment)
        if self.pyt.get('config.tf_data_dir'):
            envs['TF_DATA_DIR'] = self.pyt.get('config.tf_data_dir').format(**self.stack.data)
        #if self.pyt.get('config.tf_plugin_dir'):
        #    cli_args.append(f"-plugin-dir={self.pyt.get('config.tf_plugin_dir')}")
        if command == 'console':
            envs.update(self.get_stack_tfvariables())
            cli_args = ' '.join(self.get_stack_varfile())
        return cli_args, envs
